#include <digitalWriteFast.h>
// #include <math.h>
// #include <stdlib.h>

// Develop by Jiehan Zhang for multiplex control with Arduino Nano ESP32
// Last edit: 2024/03/21 (Unstable) (ESP32 ONLY)

// Logic: OR(serial_trigger, TTL) = HOST_order
// Pseudo-multithreading by micros():
// Thread 1: HOST_order -> check sample number -> wait sample interval -> (wait channel interval -> LED_on/cam_on -> wait exposure -> LED_off/cam_off) * channel
//           PWM mode channel loop: ((LED_on/cam_on -> wait intensity -> LED_off/cam_off) * exposure) * channel
// Thread 2~x: HOST_order -> wait stim_latency -> (wait train_interval -> (stim. on -> wait pulse period -> stim. off) * train_width)
// Only difference between 2 threads: sampling runs serially (finite state machine), but stimulation runs in parrallel across channels

// TBD:
// Sanity

// DEBUG commands:
// exposure_time,300,300,300,300,300
// sample_number,20
// sample_interval,3000
// sample_order,1,2,3,3,4
// intensity,20,40,60,80,100

// EDIT ON NEW INSTALLATION, also edit corresponding array content in following blocks
// Define pin numbers, devices with longer wavelength should go first
const int INDICATORCAM = 16;                                    // Pin number for indicator of exposure, use Arduino MEGA pin 13 (LED-L on the board), or Arduino ESP32 pin D16 (LED(B) with inversed level)
const int DEBUG = 14;											                      // Pin number for debug use
const int LED_NUM = 5;										                      // No. LED, governing size of certain arrays
const int CAM_NUM = 4;
const int STIM_NUM = 9;
const int LED[LED_NUM] = {2, 3, 4, 5, 6};                       // Example: 625 nm, 565 nm, 470 nm, 405 nm, 365 nm
const int CAM[CAM_NUM] = {7, 8, 9, 10};                         // Example: 665 nm, 600 nm, 525 nm, 470 nm
const int STIM[STIM_NUM] = {17, 18, 19, 20, 2, 3, 4, 5, 6};     // Sync. output, physical stim., or LED
const int HOST = 13;                                            // Pin number for external trigger
const int RUNNING = 11;                                         // Idle state indicator
const long int CAMPERIOD[CAM_NUM] = {20, 20, 10100, 20};        // Minimum of channel period in us. Daheng MER2-502 needs at least 34 ms (1/29.91) trigger period, and 160 needs 11 ms (1/99.45) for bin=1, Mono8
const long int CAMLATENCY = 0;                                  // WIP. For now, implement this in camera. Remeber that MER2-160 have another intrinsic 15 us, and MER2-502 have 20 us.
const int PWMFREQ = 1000;                                       // Period for PWM frequency, in us.
const int FEEDBACKSIZE = 10000;                                 // Size of the feedback arrays
const String boardID = "high_throughput_microscope";            // For recognition only
const String ERRCODE1 = "Insufficent frame interval";

// EDIT ONLY AFTER YOU`VE KNOWN
String str_event;      				                                  // Serial communication
String event_content;

// Sampling parameters
double intensity[LED_NUM] = {1, 1, 1, 1, 1};                    // 0~1
long int exposure[LED_NUM] = {5000, 5000, 5000, 5000, 5000};    // in ms
long int sample_num = 10000;
long int sample_interval = 40000;                               // in ms, time between the end and the onset of sample trains
long int chn_interval = 500;                                    // Minimum of channel interval in us. LED from Thorlabs need ~100 us to turn off.

// Stimulation parameters
// Waveform:          1 1 1    1 1 1   1 1 1   1 1 1   1 1 1   1 1 1   1 1 1
long int stim_on_latency[STIM_NUM] = {0, 9000000, 
  9000000, 9000000, 9000000, 9000000, 9000000, 
  9000000, 9000000};                                            // in ms, first blank length, typically control window 
long int stim_off_latency[STIM_NUM] = {36000000, 36000000,
  36000000, 36000000, 36000000, 36000000, 36000000, 
  36000000, 36000000};                                          // in ms, last blank onset, post-stimulation control window
long int train_interval[STIM_NUM] = {36, 9800, 
  9800, 9800, 9800, 9800, 9800, 
  9800, 9800};                                                  // in ms, second+ large blank
int train_width[STIM_NUM] = {5, 200, 
  200, 200, 200, 200, 200, 
  200, 200};                                                    // in ms, 1 1 1. Note that width shouldn`t be exactly n times of pulse period, otherwise random short pulse may be given at end.
int pulse_duration[STIM_NUM] = {4000, 1000, 
  1000, 1000, 1000, 1000, 1000, 
  1000, 1000};                                                  // in us, 1
int pulse_period[STIM_NUM] = {10000, 20000, 
  20000, 20000, 20000, 20000, 20000, 
  20000, 20000};                                                // in us, time between onset of stimulation, or 1000/frequency

// Automated control
bool HOST_order = false;                                        // external trigger listener
int i = 0;
int j = 0;

int sample_order[LED_NUM] = { -1, -1, -1, -1, -1};              // HOST interface, entries != -1 means trigger corresponding camera
bool sample_executed = false;
long int exe_count = 0;
long int sample_interval_real = 0;
unsigned long previous_micros_sample = 0;                       // Timer for sample interval
unsigned long micros_sample_chn[LED_NUM] = {0, 0, 0, 0, 0};     // Timer for sample duration (exposure)
int sample_chn = LED_NUM + 1;                                   // +1 for proper initialization
int sample_order_triggered[LED_NUM] = { -1, -1, -1, -1, -1};    // Switch to this state after triggered from HOST pin
long int channel_period[LED_NUM] = {1000, 1000, 1000, 1000, 1000};    // Actual channel period used, calculated on receiving exposure

int stim_order[STIM_NUM] = { -1, -1, 
  -1, -1, -1, -1, -1, -1, -1};
bool stim_executed[STIM_NUM] = {false, false, 
  false, false, false, false, false, false, false};
unsigned long millis_stim = 0;                                  // Timer for stimulation session
unsigned long previous_micros_stim[STIM_NUM] = {0, 0, 
  0, 0, 0, 0, 0, 0, 0};                                         // Timer for stimulation train
unsigned long current_micros_stim[STIM_NUM] = {0, 0, 
  0, 0, 0, 0, 0, 0, 0};
int stim_chn = 0;
int stim_order_triggered[STIM_NUM] = { -1, -1, 
  -1, -1, -1, -1, -1, -1, -1};                                  // Switch to this state after triggered from HOST pin

long int feedback_stim_timing[FEEDBACKSIZE];

void risingEdgeInterrupt()                  	                  // Give orders on receiving hardware TTL trigger
{
  j = 0;
  while (j < LED_NUM)
  {
    sample_order[j] = sample_order_triggered[j];                // Enable pre-defined orders
    j++;
  }
  j = 0;
  while (j < STIM_NUM)
  {
    stim_order[j] = stim_order_triggered[j];
    j++;
  }
  previous_micros_sample = 0;
}

void errHandler(int code)
{
  if (code == 1)
  {
    Serial.println(ERRCODE1);
    digitalWriteFast(INDICATORCAM, LOW);                        // Turn on 2 LED on the board
    digitalWriteFast(DEBUG, HIGH);
    delay(5000);
    ESP.restart();
  }
}

void setup()                   	 				                        // Setup pin modes
{
  pinModeFast(INDICATORCAM, OUTPUT);
  digitalWriteFast(INDICATORCAM, HIGH);
  i = 0;
  while (i < LED_NUM)
  {
    pinModeFast(LED[i++], OUTPUT);
  }
  i = 0;
  while (i < CAM_NUM)
  {
    pinModeFast(CAM[i++], OUTPUT);
  }
  i = 0;
  while (i < STIM_NUM)
  {
    pinModeFast(STIM[i++], OUTPUT);
  }
  pinModeFast(HOST, INPUT);
  pinModeFast(RUNNING, OUTPUT);
  attachInterrupt(digitalPinToInterrupt(HOST), risingEdgeInterrupt, RISING);
  Serial.begin(115200);

  while (i < FEEDBACKSIZE)                                      // Initialize the feedback arrays
  {
    feedback_stim_timing[i] = 0;
    i++;
  }
  i = 0;
}


void loop()
{
  // Structure:
  // communication();
  // if (triggered)
  // {
  //   sample();
  //   stimulation();
  // }


  // Communication interface
  // Protocol: "VARIABLE,DATA\n" or "ARRAY_VARIABLE,DATA1,DATA2,...DATAn,\n"
  if (Serial.available() > 0)
  {
    str_event = Serial.readStringUntil(',');
    if (str_event == "?")
    {
      Serial.println(boardID);
    }
    else if (str_event == "sample_number")
    {
      event_content = Serial.readStringUntil('\n');
      sample_num = event_content.toInt();
      Serial.println(str_event + sample_num);
    }
    else if (str_event == "sample_interval")
    {
      event_content = Serial.readStringUntil('\n');
      sample_interval = event_content.toDouble();
      Serial.println(str_event + sample_interval);
    }
    else if (str_event == "channel_interval")
    {
      event_content = Serial.readStringUntil('\n');
      chn_interval = event_content.toDouble();
      Serial.println(str_event + chn_interval);
    }
    else if (str_event == "intensity")
    {
      i = 0;
      while (i < LED_NUM)
      {
        event_content = Serial.readStringUntil(',');
        intensity[i] = event_content.toDouble() / 100;
        if (intensity[i] < 0.05)
        {
          intensity[i] = 0.05;
        } else if (intensity[i] > 1)
        {
          intensity[i] = 1;
        }
        i++;
      }
      event_content = Serial.readStringUntil('\n');
      Serial.println(str_event + intensity[0]);
    }
    else if (str_event == "exposure_time")
    {
      i = 0;
      while (i < LED_NUM)
      {
        event_content = Serial.readStringUntil(',');
        exposure[i++] = event_content.toDouble();
      }
      event_content = Serial.readStringUntil('\n');
      Serial.println(str_event + exposure[0]);
    }
    else if (str_event == "sample_order")
    {
      HOST_order = false;
      i = 0;
      while (i < LED_NUM)
      {
        event_content = Serial.readStringUntil(',');
        sample_order[i] = event_content.toDouble() - 1;
        if (sample_order[i] > -1)
        {
          HOST_order = true;
        }
        else
        {
          digitalWriteFast(CAM[sample_order[i]], LOW);
          digitalWriteFast(LED[i], LOW);
        }
        i++;
      }
      event_content = Serial.readStringUntil('\n');
      Serial.println(str_event + sample_order[0]);
      exe_count = 0;
    }
    else if (str_event == "sample_order_triggered")
    {
      i = 0;
      while (i < LED_NUM)
      {
        event_content = Serial.readStringUntil(',');
        sample_order_triggered[i++] = event_content.toDouble()-1;
      }
      event_content = Serial.readStringUntil('\n');
      Serial.println(str_event + sample_order_triggered[0]);
    }

    else if (str_event == "stim_on_latency")
    {
      i = 0;
      while (i < STIM_NUM)
      {
        event_content = Serial.readStringUntil(',');
        stim_on_latency[i++] = event_content.toDouble();
      }
      event_content = Serial.readStringUntil('\n');
      Serial.println(str_event + stim_on_latency[0]);
    }
    else if (str_event == "stim_off_latency")
    {
      i = 0;
      while (i < STIM_NUM)
      {
        event_content = Serial.readStringUntil(',');
        stim_off_latency[i++] = event_content.toDouble();
      }
      event_content = Serial.readStringUntil('\n');
      Serial.println(str_event + stim_off_latency[0]);
    }
    else if (str_event == "train_interval")
    {
      i = 0;
      while (i < STIM_NUM)
      {
        event_content = Serial.readStringUntil(',');
        train_interval[i++] = event_content.toDouble();
      }
      event_content = Serial.readStringUntil('\n');
      Serial.println(str_event + train_interval[0]);
    }
    else if (str_event == "train_width")
    {
      i = 0;
      while (i < STIM_NUM)
      {
        event_content = Serial.readStringUntil(',');
        train_width[i++] = event_content.toDouble();
      }
      event_content = Serial.readStringUntil('\n');
      Serial.println(str_event + train_width[0]);
    }
    else if (str_event == "pulse_duration")
    {
      i = 0;
      while (i < STIM_NUM)
      {
        event_content = Serial.readStringUntil(',');
        pulse_duration[i++] = event_content.toDouble();
      }
      event_content = Serial.readStringUntil('\n');
      Serial.println(str_event + pulse_duration[0]);
    }
    else if (str_event == "pulse_period")
    {
      i = 0;
      while (i < STIM_NUM)
      {
        event_content = Serial.readStringUntil(',');
        pulse_period[i++] = event_content.toDouble();
      }
      event_content = Serial.readStringUntil('\n');
      Serial.println(str_event + pulse_period[0]);
    }
    else if (str_event == "stim_order")
    {
      HOST_order = false;
      i = 0;
      while (i < STIM_NUM)
      {
        event_content = Serial.readStringUntil(',');
        stim_order[i] = event_content.toDouble() - 1;
        if (stim_order[i] > -1)
        {
          HOST_order = true;
        }
        else
        {
          digitalWriteFast(STIM[i], LOW);
        }
        i++;
      }
      millis_stim = millis();
      event_content = Serial.readStringUntil('\n');
      Serial.println(str_event + stim_order[0]);
    }
    else if (str_event == "stim_order_triggered")
    {
      i = 0;
      while (i < STIM_NUM)
      {
        event_content = Serial.readStringUntil(',');
        stim_order_triggered[i++] = event_content.toDouble()-1;
      }
      millis_stim = millis();
      event_content = Serial.readStringUntil('\n');
      Serial.println(str_event + stim_order_triggered[0]);
    }
    else if (str_event == "feedback_stim_timing")
    {
      // Serial.write((byte*)feedback_stim_timing, FEEDBACKSIZE * sizeof(feedback_stim_timing[0]));
      i = 0;
      while (i < FEEDBACKSIZE-1)  // Initialize the feedback arrays
      {
        Serial.print(feedback_stim_timing[i]);
        Serial.print(',');
        i++;
      }
      Serial.println(feedback_stim_timing[i]);
      i = 0;
    }

    else if (str_event == "DEBUG1")
    {
      digitalWriteFast(DEBUG, LOW);
      Serial.println(str_event);
    }
    else if (str_event == "DEBUG2")
    {
      digitalWriteFast(DEBUG, HIGH);
      Serial.println(str_event);
    }
  }


  // Listener of HOST_order
  // HOST_order = false;
  if (!HOST_order)															                // Check for orders if not running already
  {
    digitalWriteFast(RUNNING, LOW);						                  // LED indicator on hull for running state
    digitalWriteFast(INDICATORCAM, HIGH);			                  // LED indicator on chip for exposure
    i = 0;
    while (i < LED_NUM)
    {
      if (sample_order[i++] > -1)
      {
        HOST_order = true;
        break;
      }
    }
    i = 0;
    while (i < STIM_NUM)
    {
      if (stim_order[i++] > -1)
      {
        HOST_order = true;
        break;
      }
    }
    i = 0;
  }
  else															                            // After order received...
  { 
    digitalWriteFast(RUNNING, HIGH);

    // Sample loop
    if (sample_chn >= LED_NUM)								          	      // New frame if last chn sampled
    { 
      if (sample_executed == true)
      {
        exe_count++;
      }
      sample_executed = false;			                            // counter of iterator for channels
      sample_chn = 0;								                            // iterator for channels
      previous_micros_sample = micros();								        // timer for frames

      // Calculate a new frame interval as a function of exposure time and sample orders
      sample_interval_real = sample_interval;                   // Initiate interval from frame period
      i = 0;
      while (i < LED_NUM)  									    
      {
        channel_period[i] = CAMPERIOD[sample_order[i]];         // Initiate channel period from camera FPS
        channel_period[i] = max(channel_period[i], chn_interval + exposure[i]);

        if (sample_order[i] > -1)                               // Substract every valid channel
        {
          sample_interval_real -= channel_period[i];
          j = i;
        }
        i++;
      }
      if (channel_period[j] > exposure[j]) 			                // Add back the channel latency of the last channel, yielding the in-between time
      {
        sample_interval_real += channel_period[j] - exposure[j];
      }
      i = 0;

      if (sample_interval_real < 0)                             // Invalid params
      {
        errHandler(1);
      }
    }

    if (exe_count < sample_num)								          	      // so we have frames yet to sample
    {
      if (micros() - previous_micros_sample >= sample_interval_real 
		  || previous_micros_sample == 0)						                // wait after sample interval
      { 
        if (sample_order[sample_chn] > -1)					    	      // if the channel be designated to sample
        { 
          if ((micros() - micros_sample_chn[sample_chn - 1] > channel_period[sample_chn-1]) 
			  || (sample_chn == 0))							                      // wait channel delay
          {
            if (intensity[sample_chn] == 1)					    	      // 100% intensity, or...
            {
              if (digitalReadFast(CAM[sample_order[sample_chn]])// initiate exposure
				  == LOW)
              {
                digitalWriteFast(INDICATORCAM, LOW);					  // turn on indicator, LED and camera
                digitalWriteFast(LED[sample_chn], HIGH);
                digitalWriteFast(CAM[sample_order[sample_chn]], HIGH);
                micros_sample_chn[sample_chn] = micros();
              }
              else												                      // terminate exposure after time`s up
              {
                if (micros() - micros_sample_chn[sample_chn] >= exposure[sample_chn])
                {
                  digitalWriteFast(INDICATORCAM, HIGH);					// turn off indicator, LED and camera
                  digitalWriteFast(LED[sample_chn], LOW);
                  digitalWriteFast(CAM[sample_order[sample_chn]], LOW);
                  sample_chn++;
                  if (sample_chn >= LED_NUM)
                  {
                    sample_executed = true;
                  }
                }
              }
            }
            else												                        // ...PWM
            {
              if ((micros() - micros_sample_chn[sample_chn])
               % PWMFREQ < intensity[sample_chn] * PWMFREQ &&
               micros() - micros_sample_chn[sample_chn] <= 
               exposure[sample_chn] - PWMFREQ * (1 - intensity[sample_chn]))                  // PWM duty cycle
              { 
                digitalWriteFast(LED[sample_chn], HIGH);
              }
              else
              {
                digitalWriteFast(LED[sample_chn], LOW);
              }
              if (digitalReadFast(CAM[sample_order[sample_chn]])// initiate exposure
				  == LOW)
              {
                digitalWriteFast(INDICATORCAM, LOW);					  // turn on indicator and camera
                digitalWriteFast(CAM[sample_order[sample_chn]], HIGH);
                micros_sample_chn[sample_chn] = micros();
              }
              else												                      // terminate exposure after time`s up
              {
                if (micros() - micros_sample_chn[sample_chn] >= exposure[sample_chn])
                {
                  digitalWriteFast(INDICATORCAM, HIGH);					// turn of indicator, LED and camera
                  digitalWriteFast(CAM[sample_order[sample_chn]], LOW);
                  digitalWriteFast(LED[sample_chn], LOW);
                  sample_chn++;				                          // "next channel" block
                  if (sample_chn >= LED_NUM)
                  {
                    sample_executed = true;
                  }
                }
              }
            }
          }
        }
        else													                          // idle channel
        {
          digitalWriteFast(CAM[sample_order[sample_chn]], LOW);
          digitalWriteFast(LED[sample_chn], LOW);
          sample_chn++;				                                  // "next channel" block
          if (sample_chn >= LED_NUM)
          {
            sample_executed = true;
          }
        }
      }
    }
    else														                            // terminate sampling
    {
      Serial.println("sampled");
      i = 0;
      while (i < LED_NUM)
      {
        sample_order[i++] = -1;
      }
      HOST_order = false;
    }


    // Stimulation loops
    for (stim_chn = 0; stim_chn < STIM_NUM; stim_chn++)         // For each channel,
    {
      if (stim_order[stim_chn] > -1)						                // if it is designated to stimulate
      {
        if (millis() - millis_stim > stim_on_latency[stim_chn] 
         && millis() - millis_stim < stim_off_latency[stim_chn])// Start stimulation trials only within the window
        { 
          if (micros() - previous_micros_stim[stim_chn] > train_interval[stim_chn] * 1000 
            && stim_executed[stim_chn] == true )                // New train
          {
            stim_executed[stim_chn] = false;
            current_micros_stim[stim_chn] = micros();           // Timer for this train
          }

          if (stim_executed[stim_chn] == false)                 // within the train
          { 
            if ((micros() - current_micros_stim[stim_chn]) % pulse_period[stim_chn] 
              < pulse_duration[stim_chn])                       // generate pulses with certain period
            { 
              digitalWriteFast(STIM[stim_chn], HIGH);
            }
            else
            {
              digitalWriteFast(STIM[stim_chn], LOW);
            }
            if (micros() - current_micros_stim[stim_chn] 
              >= train_width[stim_chn] * 1000)                  // terminate burst after time`s up
            { 
              digitalWriteFast(STIM[stim_chn], LOW);

              stim_executed[stim_chn] = true;
              previous_micros_stim[stim_chn] = micros();        // timer for next train
            }
          }
        }
        else
        {                                                       // idle channel
          if (stim_chn >= STIM_NUM)
          {
            stim_executed[stim_chn] = true;
            previous_micros_stim[stim_chn] = current_micros_stim[stim_chn];
          }
          
        }
      }
      else
      {                                                         // idle channel
        if (stim_chn >= STIM_NUM)
        {
          stim_executed[stim_chn] = true;
          previous_micros_stim[stim_chn] = current_micros_stim[stim_chn];
        }
      }
    }
  }

  // if (digitalReadFast(LED[0]) == HIGH)
  // {
  //   digitalWriteFast(LED[0], LOW);
  // }
  // else
  // {
  //   digitalWriteFast(LED[0], HIGH);
  // }
}
